Attribute VB_Name = "Module1"
Option Explicit

Public Declare Function GetPixel Lib "gdi32" (ByVal hDC As Long, ByVal X As Long, ByVal Y As Long) As Long

Public Type Point2D
    X As Long
    Y As Long
End Type

Public Type Point2DFloat
    X As Single
    Y As Single
End Type

Public Type Polygon2D
    Vertex(0 To 2) As Point2D
End Type

Public Type Weights
    W0 As Single
    W1 As Single
    W2 As Single
End Type

Private Type RECT
    Left As Long
    Top As Long
    Right As Long
    Bottom As Long
End Type

Public Type RGB
    Red As Byte
    Green As Byte
    Blue As Byte
End Type

Public Const PI As Single = 3.14159265359

Public OnePoly As Polygon2D
Public RawPoly As Polygon2D
Public Texture() As RGB

Public Function GetRGB(Col As Long) As RGB

GetRGB.Red = (Col And &HFF&)
GetRGB.Green = (Col And &HFF00&) \ &H100
GetRGB.Blue = (Col And &HFF0000) \ &H10000
  
End Function

Private Function TriBoundingBox(ByRef Triangle As Polygon2D) As RECT

Dim TempRect As RECT

If Triangle.Vertex(0).X <= Triangle.Vertex(1).X And Triangle.Vertex(0).X <= Triangle.Vertex(2).X Then TempRect.Left = Triangle.Vertex(0).X
If Triangle.Vertex(1).X <= Triangle.Vertex(0).X And Triangle.Vertex(1).X <= Triangle.Vertex(2).X Then TempRect.Left = Triangle.Vertex(1).X
If Triangle.Vertex(2).X <= Triangle.Vertex(0).X And Triangle.Vertex(2).X <= Triangle.Vertex(1).X Then TempRect.Left = Triangle.Vertex(2).X

If Triangle.Vertex(0).X >= Triangle.Vertex(1).X And Triangle.Vertex(0).X >= Triangle.Vertex(2).X Then TempRect.Right = Triangle.Vertex(0).X
If Triangle.Vertex(1).X >= Triangle.Vertex(0).X And Triangle.Vertex(1).X >= Triangle.Vertex(2).X Then TempRect.Right = Triangle.Vertex(1).X
If Triangle.Vertex(2).X >= Triangle.Vertex(0).X And Triangle.Vertex(2).X >= Triangle.Vertex(1).X Then TempRect.Right = Triangle.Vertex(2).X

If Triangle.Vertex(0).Y <= Triangle.Vertex(1).Y And Triangle.Vertex(0).Y <= Triangle.Vertex(2).Y Then TempRect.Top = Triangle.Vertex(0).Y
If Triangle.Vertex(1).Y <= Triangle.Vertex(0).Y And Triangle.Vertex(1).Y <= Triangle.Vertex(2).Y Then TempRect.Top = Triangle.Vertex(1).Y
If Triangle.Vertex(2).Y <= Triangle.Vertex(0).Y And Triangle.Vertex(2).Y <= Triangle.Vertex(1).Y Then TempRect.Top = Triangle.Vertex(2).Y

If Triangle.Vertex(0).Y >= Triangle.Vertex(1).Y And Triangle.Vertex(0).Y >= Triangle.Vertex(2).Y Then TempRect.Bottom = Triangle.Vertex(0).Y
If Triangle.Vertex(1).Y >= Triangle.Vertex(0).Y And Triangle.Vertex(1).Y >= Triangle.Vertex(2).Y Then TempRect.Bottom = Triangle.Vertex(1).Y
If Triangle.Vertex(2).Y >= Triangle.Vertex(0).Y And Triangle.Vertex(2).Y >= Triangle.Vertex(1).Y Then TempRect.Bottom = Triangle.Vertex(2).Y

TriBoundingBox = TempRect

End Function

Private Function PointToWeights(ThisPoint As Point2D, Triangle As Polygon2D) As Weights

Dim TempWeights As Weights
Dim Denom As Single

Denom = ((Triangle.Vertex(1).Y - Triangle.Vertex(2).Y) * (Triangle.Vertex(0).X - Triangle.Vertex(2).X)) + ((Triangle.Vertex(2).X - Triangle.Vertex(1).X) * (Triangle.Vertex(0).Y - Triangle.Vertex(2).Y))

TempWeights.W0 = ((Triangle.Vertex(1).Y - Triangle.Vertex(2).Y) * (ThisPoint.X - Triangle.Vertex(2).X)) + ((Triangle.Vertex(2).X - Triangle.Vertex(1).X) * (ThisPoint.Y - Triangle.Vertex(2).Y))
TempWeights.W0 = TempWeights.W0 / Denom

TempWeights.W1 = ((Triangle.Vertex(2).Y - Triangle.Vertex(0).Y) * (ThisPoint.X - Triangle.Vertex(2).X)) + ((Triangle.Vertex(0).X - Triangle.Vertex(2).X) * (ThisPoint.Y - Triangle.Vertex(2).Y))
TempWeights.W1 = TempWeights.W1 / Denom

TempWeights.W2 = 1 - TempWeights.W0 - TempWeights.W1
        
PointToWeights = TempWeights

End Function

Private Function WeightsToPoint(TheseWeights As Weights, Triangle As Polygon2D) As Point2DFloat

Dim TempPoint As Point2DFloat

TempPoint.X = (TheseWeights.W0 * Triangle.Vertex(0).X) + (TheseWeights.W1 * Triangle.Vertex(1).X) + (TheseWeights.W2 * Triangle.Vertex(2).X)
TempPoint.Y = (TheseWeights.W0 * Triangle.Vertex(0).Y) + (TheseWeights.W1 * Triangle.Vertex(1).Y) + (TheseWeights.W2 * Triangle.Vertex(2).Y)

WeightsToPoint = TempPoint

End Function

Private Function IsPointInTriangle(ByRef CheckPoint As Point2D, ByRef Triangle As Polygon2D) As Boolean 'Quelle: https://stackoverflow.com/questions/2049582/how-to-determine-if-a-point-is-in-a-2d-triangle

Dim as_X As Long
Dim as_Y As Long
Dim s_AB As Boolean

as_X = CheckPoint.X - Triangle.Vertex(0).X
as_Y = CheckPoint.Y - Triangle.Vertex(0).Y

s_AB = ((Triangle.Vertex(1).X - Triangle.Vertex(0).X) * as_Y) - ((Triangle.Vertex(1).Y - Triangle.Vertex(0).Y) * as_X) > 0

If ((Triangle.Vertex(2).X - Triangle.Vertex(0).X) * as_Y) - ((Triangle.Vertex(2).Y - Triangle.Vertex(0).Y) * as_X) > 0 = s_AB Then
    IsPointInTriangle = False
    Exit Function
End If

If ((Triangle.Vertex(2).X - Triangle.Vertex(1).X) * (CheckPoint.Y - Triangle.Vertex(1).Y)) - ((Triangle.Vertex(2).Y - Triangle.Vertex(1).Y) * (CheckPoint.X - Triangle.Vertex(1).X)) > 0 <> s_AB Then
    IsPointInTriangle = False
    Exit Function
End If

IsPointInTriangle = True

End Function

Public Sub DrawFrame()

Dim SA As SAFEARRAY
Dim OffsetPosBase As Long
Dim Offset As Long

With SA
    .cbElements = 1
    .cDims = 1
    .Bounds(0).lLbound = 0
    .Bounds(0).cElements = Bmp.bmWidthBytes * Bmp.bmHeight
    .pvData = Bmp.bmBits
End With

Call CopyMemory(ByVal VarPtrArray(ImageBuffer), VarPtr(SA), 4)
Call CopyMemory(ImageBuffer(0), FrameBuffer(0), UBound(ImageBuffer) + 1) 'Hintergrundfarbe wiederherstellen.
Form1.Picture1.Refresh
Call CopyMemory(ByVal VarPtrArray(ImageBuffer), 0&, 4)
FrameBuffer = BackGround

End Sub

Public Sub DrawPoly(FinalPoly As Polygon2D, SourcePoly As Polygon2D)

Dim xPix As Long
Dim yPix As Long
Dim DrawX As Long
Dim DrawY As Long

Dim PolyBox As RECT
Dim MyWeights As Weights
Dim FinalPoint As Point2D

Dim SourcePoint As Point2DFloat

Dim Bilinear As Boolean

Dim X As Single
Dim Y As Single
Dim X1 As Long
Dim X2 As Long
Dim Y1 As Long
Dim Y2 As Long

Dim TL As RGB
Dim TR As RGB
Dim BL As RGB
Dim BR As RGB

Dim FinalColor As RGB

Dim WidthM1 As Long
Dim HeightM1 As Long
Dim WidthT3 As Long

Dim OffsetPosBase As Long
Dim Offset As Long

'---------------------------------------------------------------------------------

WidthM1 = IMAGEWIDTH - 1
HeightM1 = IMAGEHEIGHT - 1
WidthT3 = IMAGEWIDTH * 3

Bilinear = Form1.Check1.Value

'---------------------------------------------------------------------------------

PolyBox = TriBoundingBox(FinalPoly)

For xPix = PolyBox.Left To PolyBox.Right
    For yPix = PolyBox.Top To PolyBox.Bottom
        FinalPoint.X = xPix
        FinalPoint.Y = yPix
    
        If IsPointInTriangle(FinalPoint, FinalPoly) Then
            MyWeights = PointToWeights(FinalPoint, FinalPoly)
            SourcePoint = WeightsToPoint(MyWeights, SourcePoly)
            
            X = SourcePoint.X
            Y = SourcePoint.Y
                    
            If Bilinear Then
                X1 = Int(SourcePoint.X)
                Y1 = Int(SourcePoint.Y)
                X2 = X1 + 1
                Y2 = Y1 + 1
                
                TL = Texture(X1, Y1)
                TR = Texture(X2, Y1)
                BL = Texture(X1, Y2)
                BR = Texture(X2, Y2)
                
                FinalColor.Red = (1 / ((X2 - X1) * (Y2 - Y1))) * ((TL.Red * (X2 - X) * (Y2 - Y)) + (TR.Red * (X - X1) * (Y2 - Y)) + (BL.Red * (X2 - X) * (Y - Y1)) + (BR.Red * (X - X1) * (Y - Y1)))
                FinalColor.Green = (1 / ((X2 - X1) * (Y2 - Y1))) * ((TL.Green * (X2 - X) * (Y2 - Y)) + (TR.Green * (X - X1) * (Y2 - Y)) + (BL.Green * (X2 - X) * (Y - Y1)) + (BR.Green * (X - X1) * (Y - Y1)))
                FinalColor.Blue = (1 / ((X2 - X1) * (Y2 - Y1))) * ((TL.Blue * (X2 - X) * (Y2 - Y)) + (TR.Blue * (X - X1) * (Y2 - Y)) + (BL.Blue * (X2 - X) * (Y - Y1)) + (BR.Blue * (X - X1) * (Y - Y1)))
            Else
                FinalColor = Texture(X, Y)
            End If
            
            '-------------------------------------------------------------------------------------------------------------------------
            
            DrawX = xPix
            DrawY = yPix
            
            If DrawY < 1 Then DrawY = 1
            If DrawY > HeightM1 Then DrawY = HeightM1
            OffsetPosBase = (IMAGEHEIGHT - DrawY) * WidthT3
            
            If DrawX < 0 Then DrawX = 0
            If DrawX > WidthM1 Then DrawX = WidthM1
            Offset = OffsetPosBase + (DrawX * 3)
            
            '-------------------------------------------------------------------------------------------------------------------------
            
            FrameBuffer(Offset) = FinalColor.Blue
            FrameBuffer(Offset + 1) = FinalColor.Green
            FrameBuffer(Offset + 2) = FinalColor.Red
        End If
    Next
Next

Call DrawFrame

End Sub

Public Function RotatePoly(ThisPoly As Polygon2D, ThisRotPoint As Point2D, ByVal ThisAngle As Long) As Polygon2D

Dim MyCos As Single
Dim MySin As Single
Dim MyPoly As Polygon2D

MyPoly = ThisPoly

MyCos = Cos(ThisAngle * PI / 180)
MySin = Sin(ThisAngle * PI / 180)

'Rotation um einen bestimmten Punkt: http://www.hinterseher.de/Diplomarbeit/Transformation.html

MyPoly.Vertex(0).X = ThisRotPoint.X + (MyCos * (ThisPoly.Vertex(0).X - ThisRotPoint.X)) - (MySin * (ThisPoly.Vertex(0).Y - ThisRotPoint.Y))
MyPoly.Vertex(0).Y = ThisRotPoint.Y + (MySin * (ThisPoly.Vertex(0).X - ThisRotPoint.X)) + (MyCos * (ThisPoly.Vertex(0).Y - ThisRotPoint.Y))

MyPoly.Vertex(1).X = ThisRotPoint.X + (MyCos * (ThisPoly.Vertex(1).X - ThisRotPoint.X)) - (MySin * (ThisPoly.Vertex(1).Y - ThisRotPoint.Y))
MyPoly.Vertex(1).Y = ThisRotPoint.Y + (MySin * (ThisPoly.Vertex(1).X - ThisRotPoint.X)) + (MyCos * (ThisPoly.Vertex(1).Y - ThisRotPoint.Y))

MyPoly.Vertex(2).X = ThisRotPoint.X + (MyCos * (ThisPoly.Vertex(2).X - ThisRotPoint.X)) - (MySin * (ThisPoly.Vertex(2).Y - ThisRotPoint.Y))
MyPoly.Vertex(2).Y = ThisRotPoint.Y + (MySin * (ThisPoly.Vertex(2).X - ThisRotPoint.X)) + (MyCos * (ThisPoly.Vertex(2).Y - ThisRotPoint.Y))

RotatePoly = MyPoly

End Function

Public Sub LoadTexture()

Dim X As Long
Dim Y As Long

ReDim Texture(0 To Form1.Picture2.ScaleWidth - 1, 0 To Form1.Picture2.ScaleHeight - 1)

For X = 0 To Form1.Picture2.ScaleWidth - 1
    For Y = 0 To Form1.Picture2.ScaleHeight - 1
        Texture(X, Y) = GetRGB(GetPixel(Form1.Picture2.hDC, X, Y))
    Next
Next

End Sub
